function [W1opt,b1opt,W2opt,b2opt,Jopt,iteropt,y2opt] = mlp_desiredbp(x,d2,W10,b10,W20,b20,maxiter,nltype1,nltype2,W,indicator,figureindex,subplotindex)

% THIS FUNCTION USES A SHIFT-SCALE STRATEGY INSTEAD OF PERIODICALLY
% EXTENDING THE NONLINEARITY INVERSE!!!

% Written by Deniz Erdogmus, 10.03.2001 @ CNEL,UF
% This code is not optimized for speed etc. in any sense, use at your own risk
% PURPOSE OF THIS CODE
% trains a 2-layer MLP using the backpropagation of desired response method
% The cost function assumed is weighte MSE (multiple outputs allowed)
% J = E[(d-y)'*W*(d-y)]
% The weighting matrix W must be positive definite (and symmetric by definition), i.e. W'=W

% DECRIPTION OF INPUTS
% x: each column has a sample of the input vectors of training data
% d2: each column has a sample of the desired output vectors of training data 
% W10,b10,W20,b20: initial weights
% maxiter: maximum number of iterations allowed for training
% nltype1: nonlinearity type of 1st layer neurons
% nltype2: nonlinearity type of 2nd (output) layer neurons
% W: weighting matrix for MSE

% MLP STRUCTURAL PARAMETERS
% n0: input dimensionality
% n1: # of hidden neurons
% n2: # of outputs
% nltype1: nonlinearity of first layer		0:linear, 1:atan
% nltype2: nonlinearity of second layer	0:linear, 1:atan

% SIZES OF WEIGHT MATRICES
% W1:n1xn0
% b1:n1x1
% W2:n2xn1
% b2:n2x1

% OTHER VARIABLES USED
% d2bar: desired signal for linear output of 2nd layer
% d1: desired signal for nonlinear output of 1st layer
% d1bar: desired signal for linear output of 1st layer

% nltype: choose nonlinearity type 0:linear, 1:atan

if indicator == 0
%   maxiter = 3;	% it usually gets the best solution it can in only 1 iteration
end

N = size(d2,2);	% determine # of samples
n0 = size(x,1);n1 = size(b10,1);n2 = size(b20,1); % determine size of MLP

W1 = W10;b1 = b10;W2 = W20;b2 = b20;	% initialize weights

% evaluate performance of initial weights
[y2,dummy3,dummy1,dummy2] = mlp(x,W1,b1,W2,b2,nltype1,nltype2);clear dummy1 dummy2 dummy3;
J(1) = cost(y2,d2,W);
%figure(figureindex),clf,
%subplot(2,1,1),semilogy(0,J);drawnow;
%subplot(2,1,2),plot(d2(1,:),'r');hold on;plot(y2(1,:),'b');drawnow;

W1opt=W1;b1opt=b1;W2opt=W2;b2opt=b2;Jopt=J(1);iteropt=0;	% assign initial weights as optimal

iter = 1;
while (iter <= maxiter)
   iter
   % translate d2 tp d2bar through nonlinearity of output layer
   [dummy,d2bar] = f_finv(d2,nltype2);clear dummy;
   
   % translate d2bar to d1 (backpropagate desired signal)
   if n2<n1 % underdetermined linear system so use QR factorization
      % W2*d1=(d2bar-b2)
      [Q,R] = qr(W2');R = R(1:n2,:);
      d1 = Q*[inv(R)'*(d2bar-b2*ones(1,N));zeros(n1-n2,N)];
   elseif n2>=n1
	   d1 = inv(W2'*W*W2)*W2'*W*(d2bar-b2*ones(1,N));
   end
   
   %scale d1 to fit the range of the nonlinearity
   if nltype1 == 1	% SET FOR ONLY THIS OPTION
      T = eye(n1,n1);s = zeros(n1,1);
      for ind = 1:n1
         if max(abs(d1(ind,:))) > 1
            if std(d1(ind,:)) > 0
               T(ind,ind) = 1/(max(d1(ind,:))-min(d1(ind,:)));
               %s(ind,1) = -0.5;
            elseif std(d1(ind,:)) == 0
               s(ind,1) = -mean(d1(ind,:));
            end
         end
      end
      d1 = T*d1 + repmat(s,1,N);
	   W2 = W2*inv(T);b2 = b2-W2*inv(T)*s;
	end
   
   % translate d1 to d1bar (approximate MSE if error variance is small)
   [dummy,d1bar] = f_finv(d1,nltype1);clear dummy;
   
   % Optimize W1, b1 using least squares fit of z1 to d1bar
   if indicator == 0
      G = eye(n1);				% heuristic but works good - only one iteration
   elseif indicator == 1
      G = W2'*W*W2;				% mathematically correct approach
   end
   [W1,b1,dummy] = lssoln(x,d1bar,G,nltype1,1);clear dummy;   
   
   % Evaluate z1,y1
   [dummy1,dummy2,y1,z1] = mlp(x,W1,b1,W2,b2,nltype1,nltype2);clear dummy1 dummy2;
   
   % Optimize W2, b2 using least squares fit of z2 to d2bar
   % I have observed that if W2 is very close to its optimum,
   % the A matrix to be inverted in the following subroutine
   % becomes ill-conditioned. I will use this as a stopping 
   % criterion. If cond(A)>1e-15, then do not update W2 and W1
   % Just keep the previous values
	[W2,b2,dummy1] = lssoln(y1,d2bar,W,0,2);clear dummy1;
   
   [y2,dummy3,dummy1,dummy2] = mlp(x,W1,b1,W2,b2,nltype1,nltype2);clear dummy1 dummy2 dummy3;
   J(iter+1) = cost(y2,d2,W); % new cost
   
   if J(iter+1) <= Jopt
      Jopt = J(iter+1);		% update optimal MSE estimate
      W1opt=W1;b1opt=b1;W2opt=W2;b2opt=b2; 	% replace optimal weight estimates
   	y2opt = y2;   
      iteropt = iter;	% tracks the number of iterations required to achieve optimal solution
   end   
   
	% Display performance after the update
   if iter/1==round(iter/1) | iter==maxiter
  	   figure(figureindex),
      if indicator == 0
	     	%subplot(2,1,subplotindex),semilogy([0:iter],J);xlabel('iterations');ylabel('MSE');title('G=I');
			subplot(2,1,1),semilogy([0:iter],J);xlabel('iterations');ylabel('MSE');title('G=I');
         subplot(2,1,2),cla,plot(d2(1,:),'r');hold on;plot(y2(1,:),'b');xlabel('sample index');ylabel('desired and actual outputs');
	      drawnow;
      elseif indicator == 1
	     	subplot(2,1,1+subplotindex),semilogy([0:iter],J);xlabel('iterations');ylabel('MSE');title('G=W_2^TW_2');
        %   	subplot(2,1,2),plot(d2(1,:),'r');hold on;plot(y2(1,:),'b');xlabel('sample index');ylabel('desired and actual outputs');
	      drawnow;
      end
  	end
     
   iter = iter + 1;
end	% while iter

% Show best & test results
load mg_test.mat
[y2_test,dummy3,dummy1,dummy2] = mlp(x_test,W1opt,b1opt,W2opt,b2opt,nltype1,nltype2);
d2_test = d_test;
figure(10),clf,
for k = 1:n1
   subplot(2,3,k),plot(dummy2(k,:),dummy1(k,:),'.');
end

figure(figureindex),
subplot(3,1,1),semilogy([0:iter-1],J);xlabel('iterations');ylabel('MSE');title('G=I');
subplot(3,1,2),cla,plot(d2(1,:),'r');hold on;plot(y2opt(1,:),':b');xlabel('sample index');ylabel('desired and actual outputs');title('training')
subplot(3,1,3),cla,plot(d2_test(1,:),'r');hold on;plot(y2_test(1,:),':b');xlabel('sample index');ylabel('desired and actual outputs');title('testing');
drawnow;
training_mse = mean((d2(1,:)-y2opt(1,:)).^2)/mean(d2(1,:).^2)
testing_mse = mean((d2_test(1,:)-y2_test(1,:)).^2)/mean(d2_test(1,:).^2)


%%%%%%%%
function J = cost(y,d,W);
% evaluates the weighted MSE for a multiple input neural network
% W is the weight matrix
% J = E[(d-y)'*W*(d-y)]
N = size(y,2); % number of samples
n2 = size(y,1); % dimensionality of output
error = d-y;
Jn = 0;
Jd = 0;
for s = 1:N
   Jn = Jn + error(:,s)'*W*error(:,s);
   Jd = Jd + d(:,s)'*W*d(:,s);
end
Jn = Jn/N;
Jd = Jd/N;

J = Jn/Jd; % take the ratio of error power to desired signal power
%%%%%%%%
function [y2,z2,y1,z1] = mlp(x,W1,b1,W2,b2,nltype1,nltype2)
% evaluates the output of the MLP given its inputs and weights
N = size(x,2); % determine # of samples
z1 = W1*x+b1*ones(1,N);
[y1,dummy] = f_finv(z1,nltype1);clear dummy;
z2 = W2*y1+b2*ones(1,N);
[y2,dummy] = f_finv(z2,nltype2);clear dummy;
%%%%%%%%
function [Wopt,bopt,abort_update] = lssoln(x,d,G,nltype,layer)
% computes the least squares solutions for the weights of a linear layer
% assumes multiple inputs and single output,i.e. d has only one row of data
% nltype is the nonlinearity of the particular layer
% G is the weight matrix of the cost function - weighted MSE
fprd = fpr(d,nltype); % evaluate these for later use
n = size(x,1); % dimensionality of input vector
m = size(d,1); % dimensionality of output vector
S = size(d,2); % # of samples
A = zeros(m*n+m,m*n+m);
B = zeros(m*n+m,1);
for kr = 1:m,for lr = 1:n,
      row = (kr-1)*n+lr;
      for kc = 1:m,for lc = 1:n,
            col = (kc-1)*n+lc;
            A(row,col) = G(kc,kr)*sum(fprd(kr,:).*fprd(kc,:).*x(lr,:).*x(lc,:));
      end,end,
   	for kc = 1:m,
         col = m*n+kc;
         A(row,col) = G(kc,kr)*sum(fprd(kr,:).*fprd(kc,:).*x(lr,:));
      end
end,end,
for kr = 1:m,
   row = m*n+kr;
   for kc = 1:m,for lc = 1:n,
   		col = (kc-1)*n+lc;
         A(row,col) = G(kc,kr)*sum(fprd(kr,:).*fprd(kc,:).*x(lc,:));
   end,end,
  	for kc = 1:m,
      col = m*n+kc;
      A(row,col) = G(kc,kr)*sum(fprd(kr,:).*fprd(kc,:));
   end
end,   % finished constructing A of AX=B
for kr = 1:m,for lr = 1:n,
      row = (kr-1)*n+lr;
      B(row,1) = 0;
      for d1 = 1:m,
         B(row,1) = B(row,1) + G(d1,kr)*sum(fprd(kr,:).*fprd(d1,:).*x(lr,:).*d(d1,:));
      end
end,end,
for kr = 1:m,
   row = m*n+kr;
   B(row,1) = 0;
   for d1 = 1:m,
      B(row,1) = B(row,1) + G(d1,kr)*sum(fprd(kr,:).*fprd(d1,:).*d(d1,:));
   end
end, % finished constructing B of AX=B
X = inv(A)*B; % X=[w11...w1n...wm1...wmm,b1...bm]'
Wopt = zeros(m,n);bopt = X(m*n+1:m*n+m);
for i = 1:m,
   Wopt(i,:) = X((i-1)*n+1:(i-1)*n+n)';
end
abort_update = 0;%if layer==2 & cond(A)>1e15,abort_update = 1;else,abort_update = 0;end,
%%%%%%%%
function out = fpr(in,nltype)
% evaluates the derivative of the neuron nonlinearity at the given values
N1 = size(in,1);N2 = size(in,2);
for t1 = 1:N1
   if nltype == 0 % x
      out(t1,:) = ones(1,N2);
   elseif nltype == 1 % atan(x)
      out(t1,:) = (2/pi)./(1+in(t1,:).^2);
  elseif nltype == 2 % 1/(1+exp(-x))    
      [f,dummy1] = f_finv(in(t1,:),nltype);clear dummy1;
%      out(t1,:) = (f/2+0.5).*(0.5-f/2);
      out(t1,:) = f.*(1-f);
  end
end
%%%%%%%%
function [f,finv,range] = f_finv(in,nltype)
% nonlinearity of neurons and its inverse
N1 = size(in,1);
for t1 = 1:N1
   if nltype == 0 % x
      f(t1,:) = in(t1,:);
      finv(t1,:) = in(t1,:);
   elseif nltype == 1	% (2/pi)*atan(x)
      f(t1,:) = (2/pi)*atan(in(t1,:));
      finv(t1,:) = tan(pi/2*in(t1,:));
      range = 2;
  elseif nltype == 2    % 1/(1+exp(-x))    
%      f(t1,:) = 2*(1./(1+exp(-in(t1,:)))-0.5);
%      finv(t1,:) = -log(1./((in(t1,:)-fix(in(t1,:)))/2+0.5)-1);
      f(t1,:) = 1./(1+exp(-in(t1,:)));
      finv(t1,:) = -log(1./(in(t1,:)-floor(in(t1,:)))-1);
  end
end
%%%%%%%%

